from .most_impacting_factors_utilities import (
    numerical_types,
    bin_dataframe,
    calculate_defect_rate,
    format_number,
    generate_unique_id,
    get_columns_data,
    get_columns_statistics,
    compute_counts,
    compute_variables_impact_sql,
    handle_target_sql,
    get_table_schema_sql,
    compute_current_state_sql
)
from .custom_dates_handling import convert_date_format
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor
import dataiku
import dataikuapi
import requests
import pandas as pd
import ast
import time
import json
from webaiku.apis.dataiku.api import dataiku_api


class AnalysisChartFetcher:
    def __init__(self):
        self.client = dataiku.api_client()

        # Get project name and project
        self.project_name = dataiku_api.project_key
        self.project = self.client.get_project(self.project_name)

        # Get project variables
        self.variables = self.project.get_variables()
        self.date_column_name = self.variables["standard"]["date_column"]
        self.min_date = self.variables["standard"]["min_date"]
        self.max_date = self.variables["standard"]["max_date"]
        self.main_connection = self.variables["standard"]["main_connection"]
        self.connection_type = self.client.get_connection(
            self.main_connection
        ).get_info()["type"]

        # Input dataset informations
        self.dataset_name = "product_database_final"
        self.input_dataset = self.project.get_dataset(self.dataset_name)

        # Load the studies dataset and make sure the study_id col is a string
        self.studies_dataset = dataiku.Dataset("saved_studies")
        self.studies_df = self.studies_dataset.get_dataframe()
        self.studies_df['study_id'] = self.studies_df['study_id'].astype(str)

        # Infer schema:
        # v1.2.1: infer schema removed for now as it's confusing users
        # self.input_schema = infer_schema(self.project_name, self.main_connection, self.dataset_name)
        # Replacing by the get_schema and keeping origin_type = type for consistency with infer_schema output:
        self.input_schema = self.input_dataset.get_schema()
        for column in self.input_schema['columns']:
            column['origin_type'] = column['type']

        # Get table schema if in a supported connection:
        self.supported_connections = ["SQLServer", "Snowflake", "PostgreSQL"]

        if self.connection_type in self.supported_connections:
            # Get table name & schema:
            dataset = dataiku.Dataset(self.dataset_name)
            location_info = dataset.get_location_info()
            self.table_name = location_info.get("info", {}).get("table")
            self.table_schema = get_table_schema_sql(
                self.main_connection, self.table_name, self.input_schema
            )

        # Initialize variables
        self.analysis_results = []
        self.statistics = []
        self.selected_variables = []
        self.date_range = []
        self.target_binary = "defect_binary"
        self.global_nok_rate = None
        self.initial_nok_pct = None
        self.numerical_types = {
            "bigint",
            "int",
            "smallint",
            "tinyint",
            "double",
            "float",
            "boolean",
            "date",
        }

        self.executor = ThreadPoolExecutor()

    def get_target_variables(self):
        """
        Returns the list of possible target variables.
        """
        start_time = time.time()
        input_columns = self.input_schema["columns"]
        self.input_col_dict = {col["name"]: col for col in input_columns}

        # Creating the targets dictionary
        targets = {
            col["name"]: (
                {"type": "num"}
                if col["type"] in self.numerical_types
                else {"type": "cat"}
            )
            for col in input_columns
            if col["type"] != "date" and col["name"] != "defect_binary"
        }

        # Dates handling:
        date_min = convert_date_format(self.min_date)
        date_max = convert_date_format(self.max_date)

        date_min = date_min.split(".")[0]
        date_max = date_max.split(".")[0]

        date_min = datetime.strptime(date_min, "%Y-%m-%dT%H:%M:%S").strftime("%Y/%m/%d")
        date_max = datetime.strptime(date_max, "%Y-%m-%dT%H:%M:%S").strftime("%Y/%m/%d")

        # save informations for later:
        self.date_range = [date_min, date_max]

        end_time = time.time()
        print(
            f"[METRICS] get_target_variables computed in {round(end_time - start_time,3)} seconds "
        )
        return {"date_range": [date_min, date_max], "targets": targets}

    def get_target_range(self, dates, target):
        """
        Returns the range of the target variable, taking dates into account.
        """

        if self.input_col_dict:
            target_col = self.input_col_dict.get(target, None)
        # When working on a copy, the input_col_dict needs to be created:
        else:
            input_columns = self.input_schema["columns"]
            self.input_col_dict = {col["name"]: col for col in input_columns}

        # Initialize kwargs
        kwargs = {
            "project_name": self.project_name,
            "main_connection": self.main_connection,
            "dataset_name": self.dataset_name,
            "date_col_name": self.date_column_name,
            "col_list": [target_col],
            "aggregations": ["min", "max", "mean", "q1", "median", "q3", "stddev"],
        }

        # Date range handling:
        if dates != self.date_range:
            self.start_date = dates[0]
            self.end_date = dates[1]
            kwargs["date_from"] = self.start_date
            kwargs["date_to"] = self.end_date

        # Computing statistics:
        columns_stats = get_columns_statistics(**kwargs)

        # Formatting columns_stats values:
        if (
            target_col["type"] in numerical_types
            and int(columns_stats[target]["nb_points"]) != 0
        ):
            for col, stats in columns_stats.items():
                columns_stats[col]["min"] = format_number(float(stats["min"]))
                columns_stats[col]["max"] = format_number(float(stats["max"]))

        # Save target stats for later:
        self.target_stats = columns_stats
        return {"target": columns_stats}

    def get_target_chart_data(self, target, dates):
        start_time = time.time()

        """
        Returns all informations to create target chart
        """

        target_col = self.input_col_dict.get(target, None)

        # Initialize kwargs
        kwargs = {
            "project_name": self.project_name,
            "main_connection": self.main_connection,
            "dataset_name": self.dataset_name,
            "date_col_name": self.date_column_name,
            "col_list": [target_col],
            "aggregations": ["min", "max", "mean", "q1", "median", "q3", "stddev"],
        }

        # Date range handling:
        if dates != self.date_range:
            self.start_date = dates[0]
            self.end_date = dates[1]
            kwargs["date_from"] = self.start_date
            kwargs["date_to"] = self.end_date
        else:
            self.start_date = self.date_range[0]
            self.end_date = self.date_range[1]

        # Computing statistics:
        if not hasattr(self, "target_stats"):
            # Case of a work on a copy
            self.target_stats = get_columns_statistics(**kwargs)

        self.nb_pts_requested = int(self.target_stats[target]["nb_points"])

        # Initialize default statistics and values
        if target_col["type"] != "string":
            statistics = {
                "min": 0,
                "nbr_points": 0,
                "max": 0,
                "mean": 0,
                "lower_quartile": 0,
                "median": 0,
                "upper_quartile": 0,
                "std_dev": 0,
            }
        else:
            statistics = {"nbr_points": 0, "distinct_values": "", "mode": ""}

        values = []
        if self.nb_pts_requested > 0:
            if target_col["type"] in numerical_types:
                statistics = {
                    "min": round(float(self.target_stats[target]["min"]), 3),
                    "nbr_points": int(self.nb_pts_requested),
                    "max": round(float(self.target_stats[target]["max"]), 3),
                    "mean": round(float(self.target_stats[target]["mean"]), 3),
                    "lower_quartile": round(float(self.target_stats[target]["q1"]), 3),
                    "median": round(float(self.target_stats[target]["median"]), 3),
                    "upper_quartile": round(float(self.target_stats[target]["q3"]), 3),
                    "std_dev": round(float(self.target_stats[target]["stddev"]), 3),
                }
            else:
                statistics = {
                    "nbr_points": int(self.nb_pts_requested),
                    "distinct_values": len(self.target_stats[target]["unique_values"]),
                    "mode": str(self.target_stats[target]["mode"]),
                }

        if self.nb_pts_requested == 0:
            error_message = "no_data"

        elif self.nb_pts_requested < 50_000:
            # Requesting data
            kwargs = {
                "main_connection": self.main_connection,
                "dataset_name": self.dataset_name,
                "date_col_name": self.date_column_name,
                "col_list": [target],
                "date_from": self.start_date.replace("/", "-"),
                "date_to": self.end_date.replace("/", "-"),
            }
            target_data = get_columns_data(**kwargs)
            target_data = target_data.dropna()
            target_data = target_data.loc[target_data[target] != ""]

            # Parsing date column, saving min, max and converting to the expected format:
            target_data[self.date_column_name] = pd.to_datetime(
                target_data[self.date_column_name]
            )
            self.start_date = target_data[self.date_column_name].min()
            self.end_date = target_data[self.date_column_name].max()
            target_data[self.date_column_name] = target_data[
                self.date_column_name
            ].dt.strftime("%Y-%m-%d %H:%M:%S")
            # Saving filtered target & date variable:
            self.target_data = target_data

            if target_col["type"] != "string":
                values = [
                    {"date": row[self.date_column_name], "value": row[target]}
                    for _, row in target_data.iterrows()
                ]
            else:
                values = [
                    {"date": row[self.date_column_name], "value": str(row[target])}
                    for _, row in target_data.iterrows()
                ]

            error_message = None

        else:
            error_message = "too_large"

        response = {
            "target": target,
            "date_interval": [self.start_date, self.end_date],
            "data": values,
            "statistics": statistics,
        }
        if error_message:
            response["error_message"] = error_message

        end_time = time.time()
        print(
            f"[METRICS] {len(response['data'])} records sent. in {round(end_time - start_time,3)} seconds "
        )
        return response

    def update_target_range(self, target, good_range):
        start_time = time.time()
        """
        Returns filtered informations to update target chart
        """

        target_col = self.input_col_dict.get(target, None)

        if self.nb_pts_requested == 0:
            error_message = "no_data"
            nb_green_pts = ""
            nb_red_pts = ""
            green_data = []
            red_data = []

        elif self.nb_pts_requested < 50_000:

            if target_col["type"] in numerical_types:
                self.target_data["color"] = self.target_data[target].apply(
                    lambda x: 1 if good_range[0] <= x <= good_range[1] else 0
                )
            elif 'boolean' in target_col["type"]:
                good_range = [ast.literal_eval(s.capitalize()) for s in good_range]
                self.target_data['color'] = self.target_data[target].apply(lambda x: 1 if x in good_range else 0)
            else:
                if 'int' in target_col["origin_type"]:  # if origin is integer convert to deal with floats
                    if self.connection_type in self.supported_connections:
                        self.target_data['color'] = self.target_data[target].astype(int).apply(
                            lambda x: 1 if str(x) in good_range else 0
                            )
                    else:
                        # We need to convert the good range from a list of string to a list of int
                        good_range = [int(x) for x in good_range]

                        self.target_data['color'] = self.target_data[target].astype(int).apply(
                            lambda x: 1 if x in good_range else 0
                            )
                else:
                    self.target_data["color"] = self.target_data[target].apply(
                        lambda x: 1 if str(x) in good_range else 0
                    )

            renamed_target_data = self.target_data.rename(
                columns={self.date_column_name: "date", target: "value"}
            )

            # Separate data into green and red based on 'color' column
            green_data = (
                renamed_target_data[renamed_target_data["color"] == 1]
                .drop(columns=["color"])
                .to_dict("records")
            )
            red_data = (
                renamed_target_data[renamed_target_data["color"] == 0]
                .drop(columns=["color"])
                .to_dict("records")
            )
            error_message = ""
            nb_green_pts = len(renamed_target_data[renamed_target_data["color"] == 1])
            nb_red_pts = len(renamed_target_data[renamed_target_data["color"] == 0])

        else:
            error_message = "too_long"
            kwargs = {
                "main_connection": self.main_connection,
                "dataset_name": self.dataset_name,
                "date_col_name": self.date_column_name,
                "filter_variable": target,
                "filter_range": good_range,
            }

            if [self.start_date, self.end_date] != self.date_range:
                kwargs["date_from"] = self.start_date
                kwargs["date_to"] = self.end_date

            counts = compute_counts(**kwargs)

            nb_green_pts = counts["filtered_points"]
            nb_red_pts = int(counts["total_points"]) - int(counts["filtered_points"]) - int(counts['null_points'])
            green_data = []
            red_data = []

        if self.connection_type in self.supported_connections:
            # Asynchronous execution of handle_target_sql using ThreadPoolExecutor
            print("Launching handle_target_sql...")
            self.future = self.executor.submit(
                handle_target_sql,
                self.main_connection,
                self.table_name,
                self.table_schema,
                self.target_binary,
                target_col,
                good_range,
            )

        response = {
            "target": target,
            "error_message": error_message,
            "nb_green_points": nb_green_pts,
            "nb_red_points": nb_red_pts,
            "green_data": green_data,
            "red_data": red_data,
        }

        end_time = time.time()
        print(
            f"[METRICS] {len(response['green_data'])+len(response['red_data'])} records sent. in {round(end_time - start_time,3)} seconds "
        )

        return response

    def get_analysis_variables(self, target):

        variables = [
            col["name"]
            for col in self.input_schema["columns"]
            if col["name"] != target
            and col["type"] != "date"
            and col["name"] != "defect_binary"
        ]
        variables = sorted(variables)

        return {"target": target, "variables": variables}

    def get_analysis_results(
        self, target, num_bins, analysis_variables, dates, good_range
    ):
        """
        Returns all informations to create impact charts
        """
        start_time = time.time()
        self.selected_variables = analysis_variables
        self.target_variable = target
        self.good_range = good_range
        self.nb_bins = num_bins
        self.analysis_target_variables = analysis_variables + [self.target_variable]

        # Handling the supplied dates, useful for work on a copy option
        self.start_date = dates[0].replace("/", "-")
        self.end_date = dates[1].replace("/", "-")

        if self.connection_type in self.supported_connections:

            if not self.future.done():
                print("Waiting for self.future to complete...")
                self.future.result()
                print("self.future is now done!")

            self.global_nok_rate = self.future.result()

            # Compute results in sql:
            target_col = self.input_col_dict.get(target, None)
            impact_df, self.variable_impact = compute_variables_impact_sql(
                self.main_connection,
                self.table_name,
                self.table_schema,
                self.input_schema,
                self.global_nok_rate,
                self.target_variable,
                analysis_variables + ['defect_binary'],
                num_bins,
            )
        else:
            # Load the dataframe with the selected variables:
            kwargs = {
                "main_connection": self.main_connection,
                "dataset_name": self.dataset_name,
                "date_col_name": self.date_column_name,
                "col_list": self.analysis_target_variables,
            }

            if dates != self.date_range:
                start_date = dates[0].replace("/", "-")
                end_date = dates[1].replace("/", "-")
                kwargs["date_from"] = start_date
                kwargs["date_to"] = end_date

            get_data_start_time = time.time()
            print("Requesting data...")
            get_data_start_time = time.time()
            self.analysis_df = get_columns_data(**kwargs)
            get_data_end_time = time.time()
            self.target_data = self.analysis_df[[target]]
            print(
                f"[METRICS] get_columns_data done, it took {(get_data_end_time - get_data_start_time):.2f}s to retrieve {len(self.analysis_df)} records"
            )

            # Handle target:
            target_col = self.input_col_dict.get(target, None)
            if target_col["type"] == "string":
                # Convert all items in good_range to strings
                good_range = [str(item) for item in good_range]
                self.analysis_df[self.target_binary] = self.target_data[target].apply(
                    lambda x: "OK" if str(x) in good_range else "NOK"
                )
            else:
                lower_bound = good_range[0]
                upper_bound = good_range[1]
                self.analysis_df[self.target_binary] = self.analysis_df[target].apply(
                    lambda x: "NOK" if x < lower_bound or x > upper_bound else "OK"
                )

            # Bin the dataframe:
            bin_start_time = time.time()
            print("starting bin_dataframe...")
            df2, avg_nok_rates = bin_dataframe(
                self.analysis_df,
                self.target_binary,
                target,
                analysis_variables,
                self.input_schema,
                num_bins,
            )
            bin_end_time = time.time()
            print(f"bin_dataframe done in {(bin_end_time - bin_start_time):.2f}s!")

            # Compute impact_df and variable Correlation
            defect_start_time = time.time()
            print("starting calculate_defect_rate...")
            impact_df, self.variable_impact, _ = calculate_defect_rate(
                df2, self.date_column_name, self.target_binary, self.target_variable, avg_nok_rates
            )
            defect_end_time = time.time()
            print(
                f"calculate_defect_rate done! {(defect_end_time - defect_start_time):.2f}"
            )

        # Filter out the target_column and then select the top n variables
        top_n_variables = (
            self.variable_impact[self.variable_impact.index != target]
            .head(30)
            .index.tolist()
        )

        # Updating the selected_variables list:
        self.selected_variables = top_n_variables

        all_variables_ranges = []

        for variable in top_n_variables:

            if variable in self.variable_impact.index:
                variable_type = self.variable_impact.loc[variable]["variable_type"]
                rank = self.variable_impact.loc[variable]["rank"]
                relative_score = self.variable_impact.loc[variable]["relative_score"]
                global_nok_rate = self.variable_impact.loc[variable]["global_nok_rate"]

                # Formatting values
                avg_defect = round(global_nok_rate, 2)
                relative_score = round(relative_score, 2)

                # Getting rows from impact_df and computing average interval range:
                df_rows = impact_df[impact_df["variable"] == variable].copy()

                sig_fig = 4

                if isinstance(df_rows["bin"].iloc[0], pd.Interval):
                    # If it's a numerical variable bin, compute the sig fig
                    df_rows["interval_range"] = df_rows["bin"].apply(
                        lambda interval: interval.right - interval.left
                    )
                    avg_interval_range = (
                        df_rows.groupby("variable")["interval_range"].mean().iloc[0]
                    )

                    if avg_interval_range < 1:
                        first_row_value = df_rows["bin"].iloc[0].left
                        # length of integer part:
                        str_num = str(first_row_value)
                        parts = str_num.split(".")
                        int_part = len(parts[0])
                        dec_part = len(parts[1])

                        sig_fig = int_part + dec_part

                data = []
                for _, row in df_rows.iterrows():
                    if isinstance(row["bin"], pd.Interval):
                        # If it's a numerical variable bin
                        variable_val_min = format_number(row["bin"].left, sig_fig)
                        variable_val_max = format_number(row["bin"].right, sig_fig)
                    else:
                        # If it's a categorical variable
                        variable_val_min = str(row["bin"])
                        variable_val_max = str(row["bin"])

                    impact = format_number(row["impact"] * 100)
                    nok_rate = format_number(row["nok_rate"])
                    points_in_bin = row["points_in_bin"]
                    pct_points_in_bin = format_number(
                        100 * (points_in_bin / row["total_nb_points"])
                    )
                    color = 0 if row["color"] == "red" else 1
                    data_row = [
                        variable_val_min,
                        variable_val_max,
                        impact,
                        nok_rate,
                        points_in_bin,
                        pct_points_in_bin,
                        color,
                    ]
                    data.append(data_row)

                    # Sorting data content
                    if variable_type == "num":
                        data.sort(key=lambda row: row[0])

                    elif variable_type == "cat":
                        data.sort(
                            key=lambda x: (
                                not str(x[0]).replace(".", "", 1).isdigit(),
                                (
                                    float(x[0])
                                    if str(x[0]).replace(".", "", 1).isdigit()
                                    and str(x[0]).count(".") <= 1
                                    else x[0]
                                ),
                            )
                        )

                if data:
                    # recommended_zone = calculate_zone_impact(df_rows, global_nok_rate)
                    # Constructing JSON

                    output = {
                        "variable": variable,
                        "variable_type": variable_type,
                        "rank": int(rank),
                        "recommended_zone": [],
                        "impact": relative_score,
                        "avg_defect": avg_defect,
                        "sig_fig": sig_fig,
                        "data": data,
                    }

                    # Append the output to the list
                    all_variables_ranges.append(output)
            else:
                print(f"Variable {variable} not found in variable_impact_df")
        self.analysis_results = all_variables_ranges
        end_time = time.time()

        print(
            f"[METRICS] get_analysis_results took {round(end_time - start_time,3)} seconds."
        )

        return {"graphs": all_variables_ranges}

    def get_target_state(self, target):

        nok_count = self.variable_impact.iloc[0]["global_nok_rate"]

        if self.selected_variables:
            variables = self.selected_variables
        else:
            variables = []

        self.initial_nok_pct = nok_count
        current_points = self.nb_pts_requested
        current_nok_pct = round(100 * (nok_count / current_points), 2)

        return {
            "initial_state": {
                "points": self.nb_pts_requested,
                "nok_pct": self.initial_nok_pct,
            },
            "current_state": {"points": current_points, "nok_pct": current_nok_pct},
            "variables": variables,
        }

    def get_current_state(self, target, selected_variables_limits):

        if self.connection_type in self.supported_connections:

            # Convert selected_variables_limits to the correct format:
            filter_criteria = {}

            for item in selected_variables_limits:
                variable = item.get('variable')
                good_range = item.get('good_range')

                if variable and good_range:
                    filter_criteria[variable] = good_range

            kwargs = {
                "main_connection": self.main_connection,
                "dataset_name": self.dataset_name,
                "date_col_name": self.date_column_name,
                "filter_criteria": filter_criteria,
                "target_binary": self.target_binary,
                "table_schema": self.table_schema
            }

            if [self.start_date, self.end_date] != self.date_range:

                kwargs["date_from"] = self.start_date
                kwargs["date_to"] = self.end_date

            current_state = compute_current_state_sql(**kwargs)

            nok_count = int(current_state[2])
            current_points = int(current_state[0]) - int(current_state[1])

        else:
            state_df = self.analysis_df.copy()

            for var_limit in selected_variables_limits:
                variable = var_limit["variable"]
                good_range = var_limit["good_range"]

                if variable in self.analysis_df.columns:
                    var_type = self.input_col_dict[variable]["type"]

                    # Apply filtering based on the variable type
                    if var_type == "string":
                        state_df = state_df[state_df[variable].isin(good_range)]

                    elif var_type in self.numerical_types:
                        # For numerical variables, apply range-based filtering
                        lower_bound, upper_bound = good_range
                        state_df = state_df[
                            (state_df[variable] >= lower_bound)
                            & (state_df[variable] <= upper_bound)
                        ]

            nok_count = state_df[self.target_binary].value_counts().get("NOK", 0)
            current_points = len(state_df)

        # Check for division by zero for current_ok_pct
        if current_points == 0:
            current_nok_pct = 0
        else:
            current_nok_pct = round(100 * (nok_count / current_points), 2)

        # Check for division by zero for pfct_initial_population
        if self.nb_pts_requested == 0:
            pct_initial_population = 0
        else:
            pct_initial_population = round(
                100 * (current_points / self.nb_pts_requested), 2
            )

        # Check for division by zero for improvement_pct
        if self.initial_nok_pct == 0:
            improvement_pct = 0
        else:
            improvement_pct = round(
                ((self.initial_nok_pct - current_nok_pct) / self.initial_nok_pct) * 100,
                2,
            )

        self.statistics = {
            "target": target,
            "initial_state": {
                "points": self.nb_pts_requested,
                "nok_pct": self.initial_nok_pct,
            },
            "current_state": {
                "points": current_points,
                "nok_pct": current_nok_pct,
            },
            "ratio": {
                "pct_initial_pop": pct_initial_population,
                "pct_improvement": improvement_pct,
            },
        }

        return {
            target: target,
            "initial_state": {
                "points": self.nb_pts_requested,
                "nok_pct": self.initial_nok_pct,
            },
            "current_state": {
                "points": current_points,
                "nok_pct": current_nok_pct,
            },
            "ratio": {
                "pct_initial_pop": pct_initial_population,
                "pct_improvement": improvement_pct,
            },
        }

    def get_saved_studies(self):
        # Convert target_info from string to dictionary
        self.studies_df["target_info"] = self.studies_df["target_info"].apply(
            lambda x: ast.literal_eval(x) if isinstance(x, str) else x
        )

        # Extract target_type from target_info
        self.studies_df["target_type"] = self.studies_df["target_info"].apply(
            lambda x: x.get("target_type") if isinstance(x, dict) else None
        )

        selected_columns_df = self.studies_df[
            ["study_id", "date", "user", "target", "target_type", "date_range"]
        ]

        # Rename the DataFrame columns as per the desired output format
        renamed_studies_df = selected_columns_df.rename(
            columns={
                "study_id": "id",
            }
        )

        # Convert the DataFrame to a dictionary
        studies = renamed_studies_df.to_dict(orient="records")

        return {"studies": studies}

    def check_study_copy(self, studyId):
        """
        check saved variable ranges.
        return an error for each saved variable
        check target + variables if they still exist in dataset
        missing variable + return list of missing variable
        return specific error per check
        """

        # Loading the study:

        study = self.studies_df.loc[self.studies_df["study_id"] == studyId].iloc[0]
        study_analysis_variables = study["analysis_variables"]

        if isinstance(study_analysis_variables, str):
            study_analysis_variables = ast.literal_eval(study["analysis_variables"])

        study_target = study["target"]
        all_variables = study_analysis_variables + [study_target]

        # Checking for missing variables in the analysis:
        missing_variables = []

        for variable in all_variables:
            if variable not in self.input_col_dict:
                missing_variables.append(variable)

        # Checking for missing ranges in the variables:
        study_ranged_variables = study["variables"]

        if isinstance(study_ranged_variables, str):
            study_ranged_variables = ast.literal_eval(study["variables"])

        missing_ranges = {"cat": {}, "num": {}}
        # for each non missing variable
        for study_variable_info in study_ranged_variables:
            study_variable_name = study_variable_info["variable"]

            if study_variable_name in self.input_col_dict:
                variable_details = self.input_col_dict.get(study_variable_name, None)
                columns_stats = get_columns_statistics(
                    self.project_name,
                    self.main_connection,
                    self.dataset_name,
                    self.date_column_name,
                    [variable_details],
                )

            else:
                # Check if the variable is not in missing_variables:
                if variable not in missing_variables:

                    # For categorical variables
                    if study_variable_info["variable_type"] == "cat":
                        recommended_zone = study_variable_info["recommended_zone"]
                        target_choices = columns_stats[study_variable_name][
                            "unique_values"
                        ]

                        # Find modalities not in target_choices
                        not_in_choices = [
                            modality
                            for modality in recommended_zone
                            if modality not in target_choices
                        ]
                        missing_ranges["cat"][variable] = not_in_choices

                    # For numerical variables
                    elif study_variable_info["variable_type"] == "num":
                        min_target, max_target = (
                            columns_stats[study_variable_name]["min"],
                            columns_stats[study_variable_name]["max"],
                        )
                        out_of_range = []

                        # Check each range in 'data' to see if it's outside the target min and max
                        for data_range in study_variable_info["data"]:
                            range_min, range_max = data_range[0], data_range[1]
                            if range_min < min_target or range_max > max_target:
                                out_of_range.append([range_min, range_max])

                        missing_ranges["num"][variable] = out_of_range

        # Format the output

        output = {"missing_variables": missing_variables, "ranges_error": {}}

        # Add details for missing ranges in categorical variables
        for variable, missing_mods in missing_ranges["cat"].items():
            if missing_mods:
                output["ranges_error"][variable] = {
                    "error": "Missing Modalities",
                    "details": missing_mods,
                }

        # Add details for missing ranges in numerical variables
        for variable, missing_ranges_list in missing_ranges["num"].items():
            if missing_ranges_list:
                output["ranges_error"][variable] = {
                    "error": "Out of range",
                    "details": missing_ranges_list,
                }

        return output

    def save_study(self, input_payload, headers):
        # Reload studies
        self.studies_df = self.studies_dataset.get_dataframe()

        # Prepare target_info and add nb of bins:
        target_info = input_payload["target_info"]
        target_info["nb_bins"] = self.nb_bins

        # Prepare updated_variables:
        updated_variables = []

        # Get results
        for var_info in input_payload["variables"]:
            var_name = var_info["name"]
            var_range = var_info["range"]

            # Get analysis results for this variable
            variable_data = next(
                (
                    item
                    for item in self.analysis_results
                    if item["variable"] == var_name
                ),
                None,
            )

            # Update the recommended_zone
            if variable_data:
                variable_data["recommended_zone"] = var_range
                updated_variables.append(variable_data)
            else:
                print(f"No data found for variable {var_name}")

        if updated_variables:
            study_id = generate_unique_id(self.statistics["target"], 8)
            try:
                auth_info = dataiku.api_client().get_auth_info_from_browser_headers(
                    headers
                )
                user = auth_info["authIdentifier"]
            except (
                dataikuapi.utils.DataikuException,
                requests.exceptions.HTTPError,
            ) as e:
                print(f"Exception occurred: {str(e)}")
                user = "user_not_found"

            # Handling date range:
            target_date = target_info["target_date"]
            start_date = target_date["from"].replace("/", "-")
            end_date = target_date["to"].replace("/", "-")
            date_range = json.dumps([start_date, end_date])

            new_study = {
                "study_id": study_id,
                "date": datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
                "user": user,
                "target": self.statistics["target"],
                "target_info": target_info,
                "date_column": self.date_column_name,
                "date_range": date_range,
                "initial_state": self.statistics["initial_state"],
                "final_state": self.statistics["current_state"],
                "ratio": self.statistics["ratio"],
                "variables": updated_variables,
                "analysis_variables": input_payload["analysis_variables"],
            }

            self.studies_df = self.studies_df.append(new_study, ignore_index=True)
            self.studies_dataset.write_from_dataframe(self.studies_df)

            # Return success status and study_id
            return {"status": "OK", "study_id": str(study_id)}

    def get_study(self, study_id):
        study = self.studies_df.loc[self.studies_df["study_id"] == study_id]

        # Check if the filtered dataframe is empty
        if study.empty:
            raise ValueError("Missing study")

        # Helper function to safely evaluate a string as a literal
        def safe_literal_eval(value):
            if isinstance(value, str):
                try:
                    return ast.literal_eval(value)
                except ValueError:
                    raise ValueError("malformed node or string: " + repr(value))
            return value

        # Transform 'analysis_variables' from string to list of dictionaries and then extract 'variable' and 'recommended_zone'
        variables = safe_literal_eval(study["variables"].iloc[0])
        saved_variables = [
            {"name": item["variable"], "range": item["recommended_zone"]}
            for item in variables
        ]

        return {
            "initial_state": safe_literal_eval(study["initial_state"].iloc[0]),
            "final_state": safe_literal_eval(study["final_state"].iloc[0]),
            "ratio": safe_literal_eval(study["ratio"].iloc[0]),
            "graphs": safe_literal_eval(study["variables"].iloc[0]),
            "target_info": safe_literal_eval(study["target_info"].iloc[0]),
            "selected_variables": safe_literal_eval(
                study["analysis_variables"].iloc[0]
            ),
            "saved_variables": saved_variables,
        }

    def delete_study(self, study_id):
        # Reload the studies dataset to avoid any issue:
        self.studies_dataset = dataiku.Dataset("saved_studies")
        self.studies_df = self.studies_dataset.get_dataframe()

        indices_to_drop = self.studies_df[self.studies_df["study_id"] == study_id].index

        # Check if any indices were found
        if not indices_to_drop.empty:
            # Drop the rows by index
            self.studies_df = self.studies_df.drop(indices_to_drop)

            try:
                # Write it back to dataset
                self.studies_dataset.write_from_dataframe(self.studies_df)
                status = f"Study {study_id} deleted successfully!"
            except Exception as e:
                status = f"Failed to write the updated DataFrame back to the dataset. Error: {str(e)}"
        else:
            status = f"No study found with ID {study_id}."

        return {"status": status}
